import logging
import os
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Protocol, Sequence, Tuple, Type

from pydantic import BaseModel

from jararaca.core.providers import Token
from jararaca.microservice import (
    AppContext,
    AppInterceptor,
    AppInterceptorWithLifecycle,
    Container,
    Microservice,
)
from jararaca.tools.app_config.decorators import RequiresConfig

logger = logging.getLogger(__name__)


class ConfigParser(Protocol):

    def __call__(self, cls: Type[BaseModel]) -> BaseModel: ...


class FromEnvConfigParser(ConfigParser):

    def __call__(self, cls: Type[BaseModel]) -> BaseModel:
        return cls.model_validate({**os.environ})


class AppConfigurationInterceptor(AppInterceptor, AppInterceptorWithLifecycle):

    def __init__(
        self,
        global_configs: Sequence[Tuple[Token[Any], Type[BaseModel]]] = [],
        config_parser: ConfigParser = FromEnvConfigParser(),
    ):
        self.global_configs = global_configs
        self.config_parser = config_parser

    @asynccontextmanager
    async def intercept(self, app_context: AppContext) -> AsyncGenerator[None, None]:
        yield

    def instance_basemodels(self, basemodel_type: Type[BaseModel]) -> BaseModel:
        return self.config_parser(basemodel_type)

    def get_configs_from_application(
        self, app: Microservice, container: Container
    ) -> Sequence[Tuple[Token[Any], Type[BaseModel]]]:
        return [
            *self.global_configs,
            *[
                (config.token, config.config)
                for controller in app.controllers
                if (config := RequiresConfig.get(controller))
            ],
        ]

    @asynccontextmanager
    async def lifecycle(
        self,
        app: Microservice,
        container: Container,
    ) -> AsyncGenerator[None, None]:

        configs = [
            *self.global_configs,
            *self.get_configs_from_application(app, container),
        ]

        errors: list[str] = []

        for config in configs:
            token, basemodel_type = config

            try:
                basemodel_instance = self.instance_basemodels(basemodel_type)
            except Exception as e:
                errors.append(f"Error creating {basemodel_type}: {e}")
                continue

            container.register(basemodel_instance, token)

        if errors:
            raise AppConfigValidationError("\n".join(errors))

        yield

        logger.info("finalizando")


class AppConfigValidationError(Exception):
    pass



import inspect
import typing
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from io import StringIO
from types import NoneType, UnionType
from typing import IO, Annotated, Any, Generic, Literal, Type, TypeVar, get_origin
from uuid import UUID

from fastapi import Request, Response, UploadFile
from fastapi.params import Body, Cookie, Depends, Header, Path, Query
from fastapi.security.http import HTTPBase
from pydantic import BaseModel, PlainValidator

from jararaca.microservice import Microservice
from jararaca.presentation.decorators import HttpMapping, RestController


def snake_to_camel(snake_str: str) -> str:
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def get_field_type_for_ts(field_type: Any) -> Any:
    if field_type == UploadFile:
        return "File"
    if field_type == time:
        return "string"
    if field_type == date:
        return "string"
    if field_type == datetime:
        return "string"
    if field_type == NoneType:
        return "null"
    if field_type == UUID:
        return "string"
    if field_type == str:
        return "string"
    if field_type == int:
        return "number"
    if field_type == float:
        return "number"
    if field_type == bool:
        return "boolean"
    if field_type == Decimal:
        return "number"
    if get_origin(field_type) == tuple:
        return f"[{', '.join([get_field_type_for_ts(field) for field in field_type.__args__])}]"
    if get_origin(field_type) == list:
        return f"Array<{get_field_type_for_ts(field_type.__args__[0])}>"
    if get_origin(field_type) == set:
        return f"Array<{get_field_type_for_ts(field_type.__args__[0])}> // Set"
    if get_origin(field_type) == dict:
        return f"{{[key: {get_field_type_for_ts(field_type.__args__[0])}]: {get_field_type_for_ts(field_type.__args__[1])}}}"
    if inspect.isclass(field_type):
        if not hasattr(field_type, "__pydantic_generic_metadata__"):
            return field_type.__name__
        pydantic_metadata = getattr(field_type, "__pydantic_generic_metadata__")

        name = (
            pydantic_metadata.get("origin").__name__
            if pydantic_metadata.get("origin") is not None
            else field_type.__name__
        )
        args = pydantic_metadata.get("args")

        if len(args) > 0:
            return "%s<%s>" % (
                name,
                ", ".join([get_field_type_for_ts(arg) for arg in args]),
            )

        return name

    if hasattr(field_type, "__class__") and field_type.__class__ is TypeVar:
        return field_type.__name__
    if get_origin(field_type) == Literal:
        return " | ".join([f'"{x}"' for x in field_type.__args__])
    if get_origin(field_type) == UnionType or get_origin(field_type) == typing.Union:
        return " | ".join([get_field_type_for_ts(x) for x in field_type.__args__])
    if (get_origin(field_type) == Annotated) and (len(field_type.__args__) > 0):
        if (
            plain_validator := next(
                (x for x in field_type.__metadata__ if isinstance(x, PlainValidator)),
                None,
            )
        ) is not None:
            return get_field_type_for_ts(plain_validator.json_schema_input_type)
        return get_field_type_for_ts(field_type.__args__[0])
    return "unknown"


def is_generic_type(field_type: Any) -> bool:
    return hasattr(field_type, "__parameters__") and len(field_type.__parameters__) > 0


def get_generic_args(field_type: Any) -> Any:
    return field_type.__parameters__


def parse_type_to_typescript_interface(
    basemodel_type: Type[Any],
) -> tuple[set[type], str]:
    string_builder = StringIO()
    mapped_types: set[type] = set()

    inherited_classes = get_inherited_classes(basemodel_type)

    if (orig_type := get_origin(basemodel_type)) is not None:
        mapped_types.update(basemodel_type.__args__)
        mapped_types.add(orig_type)
        return mapped_types, ""

    if hasattr(basemodel_type, "__pydantic_generic_metadata__"):
        metadata = getattr(basemodel_type, "__pydantic_generic_metadata__")
        if metadata.get("origin") is not None:
            mapped_types.add(metadata.get("origin"))

            for arg in metadata.get("args"):
                mapped_types.update(extract_all_envolved_types(arg))

            return mapped_types, ""

    mapped_types.update(inherited_classes)

    if Enum in inherited_classes:
        return (
            set(),
            f"export enum {basemodel_type.__name__} {{\n"
            + "\n ".join([f'\t{x.value} = "{x.value}",' for x in basemodel_type])
            + "\n}\n",
        )

    valid_inherited_classes = [
        inherited_class
        for inherited_class in inherited_classes
        if inherited_class is not BaseModel
        if not is_primitive(inherited_class)
    ]

    extends_expression = (
        f" extends {', '.join([get_field_type_for_ts(inherited_class) for inherited_class in valid_inherited_classes])}"
        if len(valid_inherited_classes) > 0
        else ""
    )

    if is_generic_type(basemodel_type):
        string_builder.write(
            f"export interface {basemodel_type.__name__}<{', '.join([arg.__name__ for arg in get_generic_args(basemodel_type)])}>{extends_expression} {{\n"
        )
    else:
        string_builder.write(
            f"export interface {basemodel_type.__name__}{extends_expression} {{\n"
        )

    if hasattr(basemodel_type, "__annotations__"):
        for field_name, field in basemodel_type.__annotations__.items():
            string_builder.write(
                f"  {snake_to_camel(field_name)}: {get_field_type_for_ts(field)};\n"
            )
            mapped_types.update(extract_all_envolved_types(field))
            mapped_types.add(field)

    ## Loop over computed fields
    members = inspect.getmembers(basemodel_type, lambda a: isinstance(a, property))
    for field_name, field in members:
        if hasattr(field, "fget"):
            module_func_name = field.fget.__module__ + "." + field.fget.__qualname__
            if (
                module_func_name
                == basemodel_type.__module__
                + "."
                + basemodel_type.__name__
                + "."
                + field_name
            ):

                return_type = field.fget.__annotations__.get("return")
                if return_type is None:
                    return_type = NoneType

                string_builder.write(
                    f"  {snake_to_camel(field_name)}: {get_field_type_for_ts(return_type)};\n"
                )
                mapped_types.update(extract_all_envolved_types(return_type))
                mapped_types.add(return_type)

    string_builder.write("}\n")

    return mapped_types, string_builder.getvalue()


def get_inherited_classes(basemodel_type: Type[Any]) -> list[Type[Any]]:
    if not hasattr(basemodel_type, "__bases__"):
        return []
    return [
        base_class
        for base_class in basemodel_type.__bases__
        if get_origin(base_class) is not Generic
        if base_class is not BaseModel
    ]


def write_microservice_to_typescript_interface(
    microservice: Microservice,
) -> str:
    rest_controller_buffer = StringIO()
    mapped_types_set: set[Any] = set()

    for controller in microservice.controllers:
        rest_controller = RestController.get_controller(controller)
        if rest_controller is None:
            continue

        controller_class_str, types = write_rest_controller_to_typescript_interface(
            rest_controller, controller
        )

        mapped_types_set.update(types)
        rest_controller_buffer.write(controller_class_str)

    final_buffer = StringIO()

    final_buffer.write(
        """
export interface HttpBackendRequest {
  method: string;
  path: string;
  headers: { [key: string]: string };
  query: { [key: string]: unknown };
  body: unknown;
}

export interface HttpBackend {
  request<T>(request: HttpBackendRequest): Promise<T>;
}

export abstract class HttpService {
  constructor(protected readonly httpBackend: HttpBackend) {}
}
"""
    )

    processed_types: set[Any] = set()
    backlog: set[Any] = mapped_types_set.copy()

    unordered_types_buffer: list[tuple[type, str]] = []

    while len(backlog) > 0:
        t = backlog.pop()
        if not is_primitive(t):
            new_types, text = parse_type_to_typescript_interface(t)
            # final_buffer.write(text)
            unordered_types_buffer.append((t, text))

            backlog.update(new_types - processed_types)
            processed_types.add(t)

    ordered_types_buffer = sorted(unordered_types_buffer, key=lambda x: x[0].__name__)

    for _, text in ordered_types_buffer:
        final_buffer.write(text)

    final_buffer.write(rest_controller_buffer.getvalue())

    return final_buffer.getvalue()


def is_primitive(field_type: Any) -> bool:
    return (
        field_type
        in [
            str,
            int,
            float,
            bool,
            NoneType,
            UUID,
            Decimal,
            date,
            datetime,
            object,
            Enum,
            set,
            UploadFile,
            IO,
            time,
        ]
        or get_origin(field_type)
        in [list, dict, tuple, Literal, UnionType, Annotated, set]
        or isinstance(
            field_type,
            (
                str,
                int,
                float,
                bool,
                NoneType,
                UUID,
                Decimal,
                date,
                datetime,
                Enum,
                TypeVar,
                set,
            ),
        )
    )


def write_rest_controller_to_typescript_interface(
    rest_controller: RestController, controller: type
) -> tuple[str, set[Any]]:

    class_buffer = StringIO()

    class_buffer.write(f"export class {controller.__name__} extends HttpService {{\n")

    mapped_types: set[Any] = set()

    for name, member in inspect.getmembers(controller, predicate=inspect.isfunction):
        if (mapping := HttpMapping.get_http_mapping(member)) is not None:
            return_type = member.__annotations__.get("return")

            if return_type is None:
                return_type = NoneType

            mapped_types.update(extract_all_envolved_types(return_type))

            return_value_repr = get_field_type_for_ts(return_type)

            arg_params_spec, parametes_mapped_types = extract_parameters(
                member, rest_controller, mapping
            )

            for param in parametes_mapped_types:
                mapped_types.update(extract_all_envolved_types(param))

            class_buffer.write(
                f"\tasync {snake_to_camel(name)}({mount_parametes_arguments(arg_params_spec)}): Promise<{return_value_repr}> {{\n"
            )

            class_buffer.write("\t\t")
            class_buffer.write(
                "const response = await this.httpBackend.request<%s>({ \n"
                % return_value_repr
            )
            class_buffer.write(f'\t\t\tmethod: "{mapping.method}",\n')

            endpoint_path = parse_path_with_params(mapping.path, arg_params_spec)
            final_path = "/".join(
                s.strip("/") for s in [rest_controller.path, endpoint_path]
            )
            class_buffer.write(f"\t\t\tpath: `/{final_path}`,\n")

            class_buffer.write("\t\t\theaders: {\n")
            for param in arg_params_spec:
                if param.type_ == "header":
                    class_buffer.write(f'\t\t\t\t"{param.name}": {param.name},\n')

            class_buffer.write("\t\t\t},\n")
            class_buffer.write("\t\t\tquery: {\n")

            for param in arg_params_spec:
                if param.type_ == "query":
                    class_buffer.write(f'\t\t\t\t"{param.name}": {param.name},\n')
            class_buffer.write("\t\t\t},\n")

            if (
                body := next((x for x in arg_params_spec if x.type_ == "body"), None)
            ) is not None:
                class_buffer.write(f"\t\t\tbody: {body.name}\n")
            else:
                class_buffer.write("\t\t\tbody: undefined\n")

            class_buffer.write("\t\t});\n")
            class_buffer.write("\t\treturn response;\n")

            class_buffer.write("\t}\n")

    class_buffer.write("}\n")

    return class_buffer.getvalue(), mapped_types


EXCLUDED_REQUESTS_TYPES = [Request, Response]


@dataclass
class HttpParemeterSpec:
    type_: Literal["query", "path", "body", "header", "cookie"]
    name: str
    required: bool
    argument_type_str: str


def parse_path_with_params(path: str, parameters: list[HttpParemeterSpec]) -> str:
    for parameter in parameters:
        path = path.replace(f"{{{parameter.name}}}", f"${{{parameter.name}}}")
    return path


def mount_parametes_arguments(parameters: list[HttpParemeterSpec]) -> str:
    return ", ".join(
        [f"{parameter.name}: {parameter.argument_type_str}" for parameter in parameters]
    )


def extract_parameters(
    member: Any, controller: RestController, mapping: HttpMapping
) -> tuple[list[HttpParemeterSpec], set[Any]]:
    parameters_list: list[HttpParemeterSpec] = []
    mapped_types: set[Any] = set()
    if get_origin(member) is UnionType:
        for arg in member.__args__:
            if is_primitive(arg):
                continue
            rec_parameters, rec_mapped_types = extract_parameters(
                arg, controller, mapping
            )
            mapped_types.update(rec_mapped_types)
            parameters_list.extend(rec_parameters)
        return parameters_list, mapped_types

    if is_primitive(member):

        if get_origin(member) is Annotated:
            if (
                plain_validator := next(
                    (x for x in member.__metadata__ if isinstance(x, PlainValidator)),
                    None,
                )
            ) is not None:
                mapped_types.add(plain_validator.json_schema_input_type)
                return parameters_list, mapped_types
            return extract_parameters(member.__args__[0], controller, mapping)
        return parameters_list, mapped_types

    if hasattr(member, "__bases__"):
        for base in member.__bases__:
            # if base is not BaseModel:
            rec_parameters, rec_mapped_types = extract_parameters(
                base, controller, mapping
            )
            mapped_types.update(rec_mapped_types)
            parameters_list.extend(rec_parameters)

    if hasattr(member, "__annotations__"):
        for parameter_name, parameter_type in member.__annotations__.items():
            if parameter_name == "return":
                continue
            if parameter_type in EXCLUDED_REQUESTS_TYPES:
                continue

            if get_origin(parameter_type) == Annotated:
                annotated_type_hook = parameter_type.__metadata__[0]
                annotated_type = parameter_type.__args__[0]
                if isinstance(annotated_type_hook, Header):
                    mapped_types.add(str)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="header",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(str),
                        )
                    )
                elif isinstance(annotated_type_hook, Cookie):
                    mapped_types.add(str)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="cookie",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(str),
                        )
                    )
                elif isinstance(annotated_type_hook, Body):
                    mapped_types.update(extract_all_envolved_types(parameter_type))
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="body",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(parameter_type),
                        )
                    )
                elif isinstance(annotated_type_hook, Query):
                    mapped_types.add(parameter_type)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="query",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(parameter_type),
                        )
                    )
                elif isinstance(annotated_type_hook, Path):
                    mapped_types.add(parameter_type)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="path",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(parameter_type),
                        )
                    )

                elif isinstance(annotated_type_hook, Depends):
                    depends_hook = annotated_type_hook.dependency

                    if isinstance(depends_hook, HTTPBase):
                        ...

                    else:
                        rec_parameters, rec_mapped_types = extract_parameters(
                            depends_hook, controller, mapping
                        )
                        mapped_types.update(rec_mapped_types)
                        parameters_list.extend(rec_parameters)
                elif controller.path.find(f":{parameter_name}") != -1:
                    mapped_types.add(annotated_type)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="path",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(annotated_type),
                        )
                    )
                else:
                    mapped_types.add(annotated_type)
                    parameters_list.append(
                        HttpParemeterSpec(
                            type_="query",
                            name=parameter_name,
                            required=True,
                            argument_type_str=get_field_type_for_ts(annotated_type),
                        )
                    )

            elif inspect.isclass(parameter_type) and issubclass(
                parameter_type, BaseModel
            ):
                mapped_types.update(extract_all_envolved_types(parameter_type))
                parameters_list.append(
                    HttpParemeterSpec(
                        type_="body",
                        name=parameter_name,
                        required=True,
                        argument_type_str=get_field_type_for_ts(parameter_type),
                    )
                )
            elif (
                controller.path.find(f"{{{parameter_name}}}") != -1
                or mapping.path.find(f"{{{parameter_name}}}") != -1
            ):
                mapped_types.add(parameter_type)
                parameters_list.append(
                    HttpParemeterSpec(
                        type_="path",
                        name=parameter_name,
                        required=True,
                        argument_type_str=get_field_type_for_ts(parameter_type),
                    )
                )
            else:
                mapped_types.add(parameter_type)
                parameters_list.append(
                    HttpParemeterSpec(
                        type_="query",
                        name=parameter_name,
                        required=True,
                        argument_type_str=get_field_type_for_ts(parameter_type),
                    )
                )

            if inspect.isclass(parameter_type) and not is_primitive(parameter_type):
                signature = inspect.signature(parameter_type)

                parameter_members = signature.parameters

                for _, parameter_type in parameter_members.items():
                    if is_primitive(parameter_type.annotation):
                        if get_origin(parameter_type.annotation) is not None:
                            if (
                                get_origin(parameter_type.annotation) == Annotated
                                and (
                                    plain_validator := next(
                                        (
                                            x
                                            for x in parameter_type.annotation.__metadata__
                                            if isinstance(x, PlainValidator)
                                        ),
                                        None,
                                    )
                                )
                                is not None
                            ):
                                mapped_types.add(plain_validator.json_schema_input_type)
                            else:
                                args = parameter_type.annotation.__args__
                                mapped_types.update(args)
                        else:
                            continue
                    _, types = extract_parameters(
                        parameter_type.annotation, controller, mapping
                    )
                    mapped_types.update(types)

    if hasattr(member, "__args__"):
        for arg in member.__args__:

            rec_parameters, rec_mapped_types = extract_parameters(
                arg, controller, mapping
            )
            mapped_types.update(rec_mapped_types)
            parameters_list.extend(rec_parameters)

    return parameters_list, mapped_types


def extract_all_envolved_types(field_type: Any) -> set[Any]:
    mapped_types: set[Any] = set()

    if field_type is None:
        return mapped_types

    if is_primitive(field_type):
        if get_origin(field_type) is not None:
            if (
                get_origin(field_type) == Annotated
                and (
                    plain_validator := next(
                        (
                            x
                            for x in field_type.__metadata__
                            if isinstance(x, PlainValidator)
                        ),
                        None,
                    )
                )
                is not None
            ):
                mapped_types.add(plain_validator.json_schema_input_type)
                return mapped_types
            else:
                mapped_types.update(
                    *[extract_all_envolved_types(arg) for arg in field_type.__args__]
                )
                return mapped_types

    if inspect.isclass(field_type):
        if hasattr(field_type, "__pydantic_generic_metadata__"):
            metadata = getattr(field_type, "__pydantic_generic_metadata__")
            if metadata.get("origin") is not None:
                mapped_types.add(metadata.get("origin"))
            else:
                mapped_types.add(field_type)
            for arg in metadata.get("args"):
                mapped_types.update(extract_all_envolved_types(arg))
        else:
            mapped_types.add(field_type)

        if hasattr(field_type, "__annotations__"):
            for member in field_type.__annotations__.values():
                mapped_types.update(extract_all_envolved_types(member))

    if hasattr(field_type, "__args__"):
        for arg in field_type.__args__:
            mapped_types.update(extract_all_envolved_types(arg))

    return mapped_types



import inspect
from typing import Any, Callable, Protocol, TypeVar, cast

from fastapi import APIRouter
from fastapi import Depends as DependsF
from fastapi.exceptions import FastAPIError
from fastapi.params import Depends

from jararaca.lifecycle import AppLifecycle
from jararaca.presentation.http_microservice import HttpMiddleware

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])
DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


ControllerOptions = dict[str, Any]


class RouterFactory(Protocol):

    def __call__(self, lifecycle: AppLifecycle, instance: Any) -> APIRouter: ...


class RestController:
    REST_CONTROLLER_ATTR = "__rest_controller__"

    def __init__(
        self,
        path: str = "",
        options: ControllerOptions | None = None,
        middlewares: list[type[HttpMiddleware]] = [],
        router_factory: RouterFactory | None = None,
    ) -> None:
        self.path = path
        self.options = options
        self.router_factory = router_factory
        self.middlewares = middlewares

    def get_router_factory(
        self,
    ) -> RouterFactory:
        if self.router_factory is None:
            raise Exception("Router factory is not set")
        return self.router_factory

    def __call__(self, cls: type[DECORATED_CLASS]) -> type[DECORATED_CLASS]:

        def router_factory(
            lifecycle: AppLifecycle,
            instance: DECORATED_CLASS,
        ) -> APIRouter:
            dependencies: list[Depends] = []

            for self_middleware_type in self.middlewares:
                middleware_instance = lifecycle.container.get_by_type(
                    self_middleware_type
                )
                dependencies.append(Depends(middleware_instance.intercept))

            for middlewares_by_hook in UseMiddleware.get_middlewares(instance):
                middleware_instance = lifecycle.container.get_by_type(
                    middlewares_by_hook.middleware
                )
                dependencies.append(Depends(middleware_instance.intercept))

            for dependency in UseDependency.get_dependencies(instance):
                dependencies.append(DependsF(dependency.dependency))

            router = APIRouter(
                prefix=self.path,
                dependencies=dependencies,
                **(self.options or {}),
            )

            members = inspect.getmembers(cls, predicate=inspect.isfunction)

            router_members = [
                (name, mapping)
                for name, member in members
                if (mapping := HttpMapping.get_http_mapping(member)) is not None
            ]

            router_members.sort(key=lambda x: x[1].order)

            for name, mapping in router_members:
                route_dependencies: list[Depends] = []
                for middlewares_by_hook in UseMiddleware.get_middlewares(
                    getattr(instance, name)
                ):
                    middleware_instance = lifecycle.container.get_by_type(
                        middlewares_by_hook.middleware
                    )
                    route_dependencies.append(Depends(middleware_instance.intercept))

                for dependency in UseDependency.get_dependencies(
                    getattr(instance, name)
                ):
                    route_dependencies.append(DependsF(dependency.dependency))

                try:
                    router.add_api_route(
                        methods=[mapping.method],
                        path=mapping.path,
                        endpoint=getattr(instance, name),
                        dependencies=route_dependencies,
                        **(mapping.options or {}),
                    )
                except FastAPIError as e:
                    raise Exception(f"Error while adding route {mapping.path}") from e

            return router

        self.router_factory = router_factory

        RestController.register(cls, self)

        return cls

    @staticmethod
    def register(cls: type[DECORATED_CLASS], controller: "RestController") -> None:
        setattr(cls, RestController.REST_CONTROLLER_ATTR, controller)

    @staticmethod
    def get_controller(cls: type[DECORATED_CLASS]) -> "RestController | None":
        if not hasattr(cls, RestController.REST_CONTROLLER_ATTR):
            return None

        return cast(RestController, getattr(cls, RestController.REST_CONTROLLER_ATTR))


Options = dict[str, Any]


class HttpMapping:

    HTTP_MAPPING_ATTR = "__http_mapping__"
    ORDER_COUNTER = 0

    def __init__(
        self, method: str, path: str = "/", options: Options | None = None
    ) -> None:
        self.method = method
        self.path = path
        self.options = options

        HttpMapping.ORDER_COUNTER += 1
        self.order = HttpMapping.ORDER_COUNTER

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:

        HttpMapping.register(func, self)

        return func

    @staticmethod
    def register(func: DECORATED_FUNC, mapping: "HttpMapping") -> None:

        setattr(func, HttpMapping.HTTP_MAPPING_ATTR, mapping)

    @staticmethod
    def get_http_mapping(func: DECORATED_FUNC) -> "HttpMapping | None":

        if not hasattr(func, HttpMapping.HTTP_MAPPING_ATTR):
            return None

        return cast(HttpMapping, getattr(func, HttpMapping.HTTP_MAPPING_ATTR))


class Post(HttpMapping):

    def __init__(self, path: str = "/", options: Options | None = None) -> None:
        super().__init__("POST", path, options)


class Get(HttpMapping):

    def __init__(self, path: str = "/", options: Options | None = None) -> None:
        super().__init__("GET", path, options)


class Put(HttpMapping):

    def __init__(self, path: str = "/", options: Options | None = None) -> None:
        super().__init__("PUT", path, options)


class Delete(HttpMapping):

    def __init__(self, path: str = "/", options: Options | None = None) -> None:
        super().__init__("DELETE", path, options)


class Patch(HttpMapping):

    def __init__(self, path: str = "/", options: Options | None = None) -> None:
        super().__init__("PATCH", path, options)


class UseMiddleware:

    __MIDDLEWARES_ATTR__ = "__middlewares__"

    def __init__(self, middleware: type[HttpMiddleware]) -> None:
        self.middleware = middleware

    def __call__(self, subject: DECORATED_FUNC) -> DECORATED_FUNC:

        UseMiddleware.register(subject, self)

        return subject

    @staticmethod
    def register(subject: DECORATED_FUNC, middleware: "UseMiddleware") -> None:
        middlewares = getattr(subject, UseMiddleware.__MIDDLEWARES_ATTR__, [])
        middlewares.append(middleware)
        setattr(subject, UseMiddleware.__MIDDLEWARES_ATTR__, middlewares)

    @staticmethod
    def get_middlewares(subject: DECORATED_FUNC) -> list["UseMiddleware"]:
        return getattr(subject, UseMiddleware.__MIDDLEWARES_ATTR__, [])


class UseDependency:

    __DEPENDENCY_ATTR__ = "__dependencies__"

    def __init__(self, dependency: Any) -> None:
        self.dependency = dependency

    def __call__(self, subject: DECORATED_FUNC) -> DECORATED_FUNC:

        UseDependency.register(subject, self)

        return subject

    @staticmethod
    def register(subject: DECORATED_FUNC, dependency: "UseDependency") -> None:
        dependencies = getattr(subject, UseDependency.__DEPENDENCY_ATTR__, [])
        dependencies.append(dependency)
        setattr(subject, UseDependency.__DEPENDENCY_ATTR__, dependencies)

    @staticmethod
    def get_dependencies(subject: DECORATED_FUNC) -> list["UseDependency"]:
        return getattr(subject, UseDependency.__DEPENDENCY_ATTR__, [])



from jararaca.observability.interceptor import ObservabilityInterceptor
from jararaca.presentation.hooks import (
    raises_200_on,
    raises_400_on,
    raises_401_on,
    raises_403_on,
    raises_404_on,
    raises_422_on,
    raises_500_on,
    raises_http_exception_on,
)
from jararaca.rpc.http.decorators import Body
from jararaca.rpc.http.decorators import Delete as HttpDelete
from jararaca.rpc.http.decorators import Get as HttpGet
from jararaca.rpc.http.decorators import Header, HttpMapping
from jararaca.rpc.http.decorators import Patch as HttpPatch
from jararaca.rpc.http.decorators import PathParam
from jararaca.rpc.http.decorators import Post as HttpPost
from jararaca.rpc.http.decorators import Put as HttpPut
from jararaca.rpc.http.decorators import Query, RequestAttribute, RestClient

from .core.providers import ProviderSpec, Token
from .di import Container
from .messagebus import Message
from .messagebus.decorators import IncomingHandler, MessageBusController
from .messagebus.interceptors.publisher_interceptor import (
    AIOPikaConnectionFactory,
    MessageBusPublisherInterceptor,
)
from .messagebus.publisher import use_publisher
from .messagebus.worker import MessageBusWorker
from .microservice import Microservice, use_current_container
from .persistence.base import (
    T_BASEMODEL,
    BaseEntity,
    CriteriaBasedAttributeQueryInjector,
    CRUDOperations,
    DateCriteria,
    DatedEntity,
    DateOrderedFilter,
    DateOrderedQueryInjector,
    Identifiable,
    IdentifiableEntity,
    Paginated,
    PaginatedFilter,
    PaginatedQueryInjector,
    QueryInjector,
    QueryOperations,
    StringCriteria,
)
from .persistence.interceptors.aiosqa_interceptor import (
    AIOSQAConfig,
    AIOSqlAlchemySessionInterceptor,
    use_session,
)
from .presentation.decorators import (
    Delete,
    Get,
    Patch,
    Post,
    Put,
    RestController,
    UseDependency,
    UseMiddleware,
)
from .presentation.http_microservice import HttpMicroservice, HttpMiddleware
from .presentation.server import create_http_server
from .presentation.websocket.decorators import WebSocketEndpoint
from .presentation.websocket.redis import RedisWebSocketConnectionBackend
from .presentation.websocket.websocket_interceptor import (
    WebSocketInterceptor,
    use_ws_manager,
)
from .scheduler.decorators import ScheduledAction
from .tools.app_config.interceptor import AppConfigurationInterceptor

__all__ = [
    "raises_http_exception_on",
    "raises_200_on",
    "raises_422_on",
    "raises_404_on",
    "raises_400_on",
    "raises_401_on",
    "raises_403_on",
    "raises_500_on",
    "HttpMiddleware",
    "HttpMapping",
    "RequestAttribute",
    "Body",
    "Query",
    "Header",
    "PathParam",
    "RestClient",
    "HttpPost",
    "HttpGet",
    "HttpPatch",
    "HttpPut",
    "HttpDelete",
    "ObservabilityInterceptor",
    "QueryInjector",
    "HttpMicroservice",
    "use_current_container",
    "T_BASEMODEL",
    "DatedEntity",
    "BaseEntity",
    "use_ws_manager",
    "WebSocketEndpoint",
    "CriteriaBasedAttributeQueryInjector",
    "Identifiable",
    "IdentifiableEntity",
    "Message",
    "StringCriteria",
    "DateCriteria",
    "DateOrderedFilter",
    "DateOrderedQueryInjector",
    "Paginated",
    "PaginatedFilter",
    "PaginatedQueryInjector",
    "QueryOperations",
    "CRUDOperations",
    "RestController",
    "MessageBusController",
    "IncomingHandler",
    "ScheduledAction",
    "Microservice",
    "ProviderSpec",
    "Token",
    "AIOSqlAlchemySessionInterceptor",
    "AIOSQAConfig",
    "create_http_server",
    "MessageBusWorker",
    "Container",
    "WebSocketInterceptor",
    "use_session",
    "Post",
    "Get",
    "Patch",
    "Put",
    "Delete",
    "use_publisher",
    "AIOPikaConnectionFactory",
    "MessageBusPublisherInterceptor",
    "RedisWebSocketConnectionBackend",
    "AppConfigurationInterceptor",
    "UseMiddleware",
    "UseDependency",
]
